@@ -1,20 +1,22 @@
 package Catalyst::Model::DBIC::Schema;
 
-use strict;
-use warnings;
-
-our $VERSION = '0.23';
-
-use base qw/Catalyst::Model Class::Accessor::Fast Class::Data::Accessor/;
-use MRO::Compat;
+use Moose;
 use mro 'c3';
-use UNIVERSAL::require;
-use Carp;
-use Data::Dumper;
-require DBIx::Class;
+extends 'Catalyst::Model';
+with 'CatalystX::Component::Traits';
 
-__PACKAGE__->mk_classaccessor('composed_schema');
-__PACKAGE__->mk_accessors('schema');
+our $VERSION = '0.24';
+
+use namespace::autoclean;
+use Carp::Clan '^Catalyst::Model::DBIC::Schema';
+use Data::Dumper;
+use DBIx::Class ();
+use Moose::Autobox;
+
+use Catalyst::Model::DBIC::Schema::Types
+    qw/ConnectInfo SchemaClass CursorClass/;
+
+use MooseX::Types::Moose qw/ArrayRef Str ClassName Undef/;
 
 =head1 NAME
 
@@ -71,12 +73,11 @@ MyApp/Model/FilmDB.pm:
 
   __PACKAGE__->config(
       schema_class => 'MyApp::Schema::FilmDB',
-      connect_info => [
-                        "DBI:...",
-                        "username",
-                        "password",
-                        {AutoCommit => 1}
-                      ]
+      connect_info => {
+                        dsn => "DBI:...",
+                        user => "username",
+                        password => "password",
+                      }
   );
 
 See below for a full list of the possible config parameters.
@@ -89,19 +90,34 @@ be used/accessed in the normal Catalyst manner, via $c->model():
   my $actor = $c->model('FilmDB::Actor')->find(1);
 
 You can also use it to set up DBIC authentication with 
-Authentication::Store::DBIC in MyApp.pm:
+L<Catalyst::Authentication::Store::DBIx::Class> in MyApp.pm:
 
   package MyApp;
 
-  use Catalyst qw/... Authentication::Store::DBIC/;
+  use Catalyst qw/... Authentication .../;
 
   ...
 
-  __PACKAGE__->config->{authentication}{dbic} = {
-      user_class      => 'FilmDB::Actor',
-      user_field      => 'name',
-      password_field  => 'password'
-  }
+  __PACKAGE__->config->{authentication} = 
+                {  
+                    default_realm => 'members',
+                    realms => {
+                        members => {
+                            credential => {
+                                class => 'Password',
+                                password_field => 'password',
+                                password_type => 'hashed'
+                                password_hash_type => 'SHA-256'
+                            },
+                            store => {
+                                class => 'DBIx::Class',
+                                user_model => 'DB::User',
+                                role_relation => 'roles',
+                                role_field => 'rolename',                   
+                            }
+                        }
+                    }
+                };
 
 C<< $c->model('Schema::Source') >> returns a L<DBIx::Class::ResultSet> for 
 the source name parameter passed. To find out more about which methods can 
@@ -183,15 +199,13 @@ for more info.
 
 =head1 CONFIG PARAMETERS
 
-=over 4
-
-=item schema_class
+=head2 schema_class
 
 This is the classname of your L<DBIx::Class::Schema> Schema.  It needs
 to be findable in C<@INC>, but it does not need to be inside the 
 C<Catalyst::Model::> namespace.  This parameter is required.
 
-=item connect_info
+=head2 connect_info
 
 This is an arrayref of connection parameters, which are specific to your
 C<storage_type> (see your storage type documentation for more details). 
@@ -210,39 +224,41 @@ of the arguments supported.
 
 Examples:
 
-  connect_info => [ 'dbi:Pg:dbname=mypgdb', 'postgres', '' ],
+  connect_info => {
+    dsn => 'dbi:Pg:dbname=mypgdb',
+    user => 'postgres',
+    password => ''
+  }
 
-  connect_info => [
-                    'dbi:SQLite:dbname=foo.db',
-                    {
-                      on_connect_do => [
-                        'PRAGMA synchronous = OFF',
-                      ],
-                    }
-                  ],
+  connect_info => {
+    dsn => 'dbi:SQLite:dbname=foo.db',
+    on_connect_do => [
+      'PRAGMA synchronous = OFF',
+    ]
+  }
 
-  connect_info => [
-                    'dbi:Pg:dbname=mypgdb',
-                    'postgres',
-                    '',
-                    { AutoCommit => 0 },
-                    {
-                      on_connect_do => [
-                        'some SQL statement',
-                        'another SQL statement',
-                      ],
-                    }
-                  ],
+  connect_info => {
+    dsn => 'dbi:Pg:dbname=mypgdb',
+    user => 'postgres',
+    password => '',
+    pg_enable_utf8 => 1,
+    on_connect_do => [
+      'some SQL statement',
+      'another SQL statement',
+    ],
+  }
 
 Or using L<Config::General>:
 
     <Model::FilmDB>
         schema_class   MyApp::Schema::FilmDB
-        connect_info   dbi:Pg:dbname=mypgdb
-        connect_info   postgres
-        connect_info
+        traits Caching
         <connect_info>
-            AutoCommit   0
+            dsn   dbi:Pg:dbname=mypgdb
+            user   postgres
+            password ""
+            auto_savepoint 1
+	    quote_char """
             on_connect_do   some SQL statement
             on_connect_do   another SQL statement
         </connect_info>
@@ -255,118 +271,228 @@ or
         connect_info   dbi:SQLite:dbname=foo.db
     </Model::FilmDB>
 
+Or using L<YAML>:
 
-=item storage_type
+  Model::MyDB:
+      schema_class: MyDB
+      connect_info:
+          dsn: dbi:Oracle:mydb
+          user: mtfnpy
+          password: mypass
+          LongReadLen: 1000000
+          LongTruncOk: 1
+          on_connect_do: [ "alter session set nls_date_format = 'YYYY-MM-DD HH24:MI:SS'" ]
+          cursor_class: 'DBIx::Class::Cursor::Cached'
+	  quote_char: '"'
+
+The old arrayref style with hashrefs for L<DBI> then L<DBIx::Class> options is also
+supported:
+
+  connect_info => [
+    'dbi:Pg:dbname=mypgdb',
+    'postgres',
+    '',
+    {
+      pg_enable_utf8 => 1,
+    },
+    {
+      auto_savepoint => 1,
+      on_connect_do => [
+        'some SQL statement',
+        'another SQL statement',
+      ],
+    }
+  ]
+
+=head2 traits
+
+Array of Traits to apply to the instance. Traits are L<Moose::Role>s.
+
+They are relative to the C<< MyApp::TraitFor::Model::DBIC::Schema:: >>, then the C<<
+Catalyst::TraitFor::Model::DBIC::Schema:: >> namespaces, unless prefixed with C<+>
+in which case they are taken to be a fully qualified name. E.g.:
+
+    traits Caching
+    traits +MyApp::TraitFor::Model::Foo
+
+A new instance is created at application time, so any consumed required
+attributes, coercions and modifiers will work.
+
+Traits are applied at L<Catalyst::Component/COMPONENT> time using
+L<CatalystX::Component::Traits>.
+
+C<ref $self> will be an anon class if any traits are applied, C<<
+$self->_original_class_name >> will be the original class.
+
+When writing a Trait, interesting points to modify are C<BUILD>, L</setup> and
+L</ACCEPT_CONTEXT>.
+
+Traits that come with the distribution:
+
+=over 4
+
+=item L<Catalyst::TraitFor::Model::DBIC::Schema::Caching>
+
+=item L<Catalyst::TraitFor::Model::DBIC::Schema::Replicated>
+
+=back
+
+=head2 storage_type
 
 Allows the use of a different C<storage_type> than what is set in your
 C<schema_class> (which in turn defaults to C<::DBI> if not set in current
 L<DBIx::Class>).  Completely optional, and probably unnecessary for most
 people until other storage backends become available for L<DBIx::Class>.
 
-=back
+=head1 ATTRIBUTES
+
+The keys you pass in the model configuration are available as attributes.
+
+Other attributes available:
+
+=head2 connect_info
+
+Your connect_info args normalized to hashref form (with dsn/user/password.) See
+L<DBIx::Class::Storage::DBI/connect_info> for more info on the hashref form of
+L</connect_info>.
+
+=head2 model_name
+
+The model name L<Catalyst> uses to resolve this model, the part after
+C<::Model::> or C<::M::> in your class name. E.g. if your class name is
+C<MyApp::Model::DB> the L</model_name> will be C<DB>.
+
+=head2 _default_cursor_class
+
+What to reset your L<DBIx::Class::Storage::DBI/cursor_class> to if a custom one
+doesn't work out. Defaults to L<DBIx::Class::Storage::DBI::Cursor>.
+
+=head1 ATTRIBUTES FROM L<MooseX::Traits::Pluggable>
+
+=head2 _original_class_name
+
+The class name of your model before any L</traits> are applied. E.g.
+C<MyApp::Model::DB>.
+
+=head2 _traits
+
+Unresolved arrayref of traits passed in the config.
+
+=head2 _resolved_traits
+
+Traits you used resolved to full class names.
 
 =head1 METHODS
 
-=over 4
-
-=item new
+=head2 new
 
 Instantiates the Model based on the above-documented ->config parameters.
 The only required parameter is C<schema_class>.  C<connect_info> is
 required in the case that C<schema_class> does not already have connection
 information defined for it.
 
-=item schema
+=head2 schema
 
 Accessor which returns the connected schema being used by the this model.
 There are direct shortcuts on the model class itself for
 schema->resultset, schema->source, and schema->class.
 
-=item composed_schema
+=head2 composed_schema
 
 Accessor which returns the composed schema, which has no connection info,
 which was used in constructing the C<schema> above.  Useful for creating
 new connections based on the same schema/model.  There are direct shortcuts
 from the model object for composed_schema->clone and composed_schema->connect
 
-=item clone
+=head2 clone
 
 Shortcut for ->composed_schema->clone
 
-=item connect
+=head2 connect
 
 Shortcut for ->composed_schema->connect
 
-=item source
+=head2 source
 
 Shortcut for ->schema->source
 
-=item class
+=head2 class
 
 Shortcut for ->schema->class
 
-=item resultset
+=head2 resultset
 
 Shortcut for ->schema->resultset
 
-=item storage
+=head2 storage
 
 Provides an accessor for the connected schema's storage object.
 Used often for debugging and controlling transactions.
 
-=back
-
 =cut
 
-sub new {
-    my $self = shift->next::method(@_);
-    
-    my $class = ref($self);
-    my $model_name = $class;
-    $model_name =~ s/^[\w:]+::(?:Model|M):://;
+has schema => (is => 'rw', isa => 'DBIx::Class::Schema');
 
-    croak "->config->{schema_class} must be defined for this model"
-        unless $self->{schema_class};
+has schema_class => (
+    is => 'ro',
+    isa => SchemaClass,
+    coerce => 1,
+    required => 1
+);
 
-    my $schema_class = $self->{schema_class};
+has storage_type => (is => 'rw', isa => Str);
 
-    $schema_class->require
-        or croak "Cannot load schema class '$schema_class': $@";
+has connect_info => (is => 'ro', isa => ConnectInfo, coerce => 1);
 
-    if( !$self->{connect_info} ) {
+has model_name => (
+    is => 'ro',
+    isa => Str,
+    required => 1,
+    lazy_build => 1,
+);
+
+has _default_cursor_class => (
+    is => 'ro',
+    isa => CursorClass,
+    default => 'DBIx::Class::Storage::DBI::Cursor',
+    coerce => 1
+);
+
+sub BUILD {
+    my $self = shift;
+    my $class = $self->_original_class_name;
+    my $schema_class = $self->schema_class;
+
+    if( !$self->connect_info ) {
         if($schema_class->storage && $schema_class->storage->connect_info) {
-            $self->{connect_info} = $schema_class->storage->connect_info;
+            $self->connect_info($schema_class->storage->connect_info);
         }
         else {
-            croak "Either ->config->{connect_info} must be defined for $class"
+            die "Either ->config->{connect_info} must be defined for $class"
                   . " or $schema_class must have connect info defined on it."
 		  . " Here's what we got:\n"
 		  . Dumper($self);
         }
     }
 
-    $self->composed_schema($schema_class->compose_namespace($class));
-    $self->schema($self->composed_schema->clone);
-
-    $self->schema->storage_type($self->{storage_type})
-        if $self->{storage_type};
-
-    $self->schema->connection( 
-        ref $self->{connect_info} eq 'ARRAY' ? 
-        @{$self->{connect_info}} : 
-        $self->{connect_info}
-    );
-    
-    no strict 'refs';
-    foreach my $moniker ($self->schema->sources) {
-        my $classname = "${class}::$moniker";
-        *{"${classname}::ACCEPT_CONTEXT"} = sub {
-            shift;
-            shift->model($model_name)->resultset($moniker);
-        }
+    if (exists $self->connect_info->{cursor_class}) {
+        eval { Class::MOP::load_class($self->connect_info->{cursor_class}) }
+            or croak "invalid connect_info: Cannot load your cursor_class"
+        . " ".$self->connect_info->{cursor_class}.": $@";
     }
 
-    return $self;
+    $self->setup;
+
+    $self->composed_schema($schema_class->compose_namespace($class));
+
+    $self->schema($self->composed_schema->clone);
+
+    $self->schema->storage_type($self->storage_type)
+        if $self->storage_type;
+
+    $self->schema->connection($self->connect_info);
+
+    $self->_install_rs_models;
 }
 
 sub clone { shift->composed_schema->clone(@_); }
@@ -375,6 +501,80 @@ sub connect { shift->composed_schema->connect(@_); }
 
 sub storage { shift->schema->storage(@_); }
 
+sub resultset { shift->schema->resultset(@_); }
+
+=head2 setup
+
+Called at C<BUILD>> time before configuration, but after L</connect_info> is
+set. To do something after configuuration use C<< after BUILD => >>.
+
+=cut
+
+sub setup { 1 }
+
+=head2 ACCEPT_CONTEXT
+
+Point of extension for doing things at C<< $c->model >> time with context,
+returns the model instance, see L<Catalyst::Manual::Intro/ACCEPT_CONTEXT> for
+more information.
+
+=cut
+
+sub ACCEPT_CONTEXT { shift }
+
+sub _install_rs_models {
+    my $self  = shift;
+    my $class = $self->_original_class_name;
+
+    no strict 'refs';
+
+    my @sources = $self->schema->sources;
+
+    die "No sources found (did you forget to define your tables?)"
+        unless @sources;
+
+    foreach my $moniker (@sources) {
+        my $classname = "${class}::$moniker";
+        *{"${classname}::ACCEPT_CONTEXT"} = sub {
+            shift;
+            shift->model($self->model_name)->resultset($moniker);
+        }
+    }
+}
+
+sub _reset_cursor_class {
+    my $self = shift;
+
+    if ($self->storage->can('cursor_class')) {
+	$self->storage->cursor_class($self->_default_cursor_class)
+	    if $self->storage->cursor_class ne $self->_default_cursor_class;
+    }
+}
+
+{
+    my %COMPOSED_CACHE;
+
+    sub composed_schema {
+	my $self = shift;
+	my $class = $self->_original_class_name;
+	my $store = \$COMPOSED_CACHE{$class}{$self->schema_class};
+
+	$$store = shift if @_;
+
+	return $$store
+    }
+}
+
+sub _build_model_name {
+    my $self  = shift;
+    my $class = $self->_original_class_name;
+    (my $model_name = $class) =~ s/^[\w:]+::(?:Model|M):://;
+
+    return $model_name;
+}
+
+__PACKAGE__->meta->make_immutable;
+
 =head1 SEE ALSO
 
 General Catalyst Stuff:
@@ -385,11 +585,21 @@ L<Catalyst::Response>, L<Catalyst::Helper>, L<Catalyst>,
 Stuff related to DBIC and this Model style:
 
 L<DBIx::Class>, L<DBIx::Class::Schema>,
-L<DBIx::Class::Schema::Loader>, L<Catalyst::Helper::Model::DBIC::Schema>
+L<DBIx::Class::Schema::Loader>, L<Catalyst::Helper::Model::DBIC::Schema>,
+L<MooseX::Object::Pluggable>
+
+Traits:
+
+L<Catalyst::TraitFor::Model::DBIC::Schema::Caching>,
+L<Catalyst::TraitFor::Model::DBIC::Schema::Replicated>
 
 =head1 AUTHOR
 
-Brandon L Black, C<blblack@gmail.com>
+Brandon L Black, C<blblack at gmail.com>
+
+Contributors:
+
+Rafael Kitover, C<rkitover at cpan.org>
 
 =head1 COPYRIGHT
 
@@ -399,3 +609,4 @@ under the same terms as Perl itself.
 =cut
 
 1;
+# vim:sts=4 sw=4 et:
